'' VGA_1024.spin
'' MODIFED BY JEFF LEDGER / AKA OLDBITCOLLECTOR

{{
Info            : VT100 Terminal for AVR CP/M
Autor           : Joe G.
Version         : 00
Subversion      : 06
Funktion        : Keyboard Driver DE, VGA-1024x768 Driver, V24 Driver, VT100 Terminal
Komponenten     : vga_Hires_Text

Log             

10.04.2013      Start, CR, LF
12.04.2013      code CA, BS
17.04.2013      fast insert line
26.04.2013      arrow keys, cursor (not) visible
27.04.2013      bug in ESC[K
25.05.2013      bell
03.11.2013      add two lowres VGA modes
03.11.2013      modify beep output
}}


CON
  cols       = 80                 'number of screen columns
  rows       = 28                 'number of screen rows
  chars      = rows*cols          'number of screen characters
  esc        = $CB                'keyboard esc char
  CP_M       = 0                  'CP/M Reset
  def_cursor = %001               'block
' def_cursor = %110               'undersore, slow blink  


OBJ
 vga : "vga_Hires_Text_010"
 spk : "E555_SPKEngine.spin"

 
VAR
  byte  screen[chars]           ' screen character buffer
  word  colors[rows]            ' color specs for each screen row (see ColorPtr description above)
  byte  cursor[6]               ' cursor info array (see CursorPtr description above)
  byte  save_cursor[6]          ' cursor info array (see CursorPtr description above)
  long  sync, loc, xloc, yloc   ' sync used by VGA routine, others are local screen pointers
  long  kbdreq                  ' global val of kbdflag
  byte  inverse
  byte  bell                    ' Beeper pin
  byte  spk_engine              ' use spk engine

  
PUB start(BasePin, BeepPin, UseSPKEngine) | i, char
''
'' initialisation for Beeper/Buzzer
  bell       := BeepPin
  spk_engine := UseSPKEngine


''start vga
  vga.start(BasePin, @screen, @colors, @cursor, @sync)
  waitcnt(clkfreq * 0.3 + cnt)    'wait .3 second for cogs to start

''init screen colors to gold on blue
  repeat i from 0 to rows - 1
    colors[i] := $08F0  '08F0        '$2804 (if you want cyan on blue)

''init cursor attributes
  cursor[2] := def_cursor       'init cursor 0 to underscore with slow blink
  inverse := 0                  'not inverse character
  cls                           'clear screen

  dira[bell] := 1                                       'output
  dira[CP_M] := 0                                       'input

  
PUB beep

  ifnot spk_engine
    outa[bell] := 1                   'set bell high
    waitcnt(clkfreq/8 + cnt)          'wait a moment
    outa[bell] := 0                   'set bell low 
  else
    spk.speakerVolume( 100, bell)
    spk.speakerFrequency( 444, bell)
    waitcnt(clkfreq/8 + cnt)          'wait a moment
    spk.speakerVolume( -1, bell)


PUB CP_M_Reset
  dira[CP_M] := 1                                       'output   
  outa[CP_M] := 0        'set CP/M reset low 
  waitcnt(clkfreq/128 + cnt)'wait a moment
  outa[CP_M] := 1        'set CP/M reset high
  dira[CP_M] := 0                                       'input   

  
PUB bin(value, digits)
'' Print a binary number, specify number of digits

  repeat while digits > 32
    out("0")
    digits--

  value <<= 32 - digits

  repeat digits
    out((value <-= 1) & 1 + "0")

  
PUB binFP(value) | bitnum, bit, bitval
'' Prints FP long in special Binary format: sign, exp, mantissa

  repeat bitnum from 31 to 0
    bit := 1 << bitnum                ' create mask bit
    bitval := (bit & value) >> bitnum  ' extract bit and shift back to bit 0

    bin(bitval, 1)                    ' display one bit

    case bitnum
      27,20,16,12,8,4: out($20)       ' space after every 4 in group
      31,23: str(string("  "))        ' two after sign and exponent

  
PUB cls
'' clear screen, fill buffer with blank 

  longfill(@screen, $20202020, chars/4)

  xloc := cursor[0] := 0 'right & left
  yloc := cursor[1] := 0 'from top
  loc  := xloc + yloc*cols


PUB home

  xloc := cursor[0] := 0 'right & left
  yloc := cursor[1] := 0 'from top
  loc  := xloc + yloc*cols

  
PUB color(ColorVal) | i
''reset screen colors
  repeat i from 0 to rows - 1
    colors[i] := $0000 + ColorVal         

  
PUB rowcolor(ColorVal, row) 
'' reset row color to colorval
  if row > rows-1
    row := rows-1

   colors[row] := $0000 + ColorVal


PUB cursloc(x, y)

'' move cursor to x, y position
  if y < 0
    y := 0
  if x < 0
    x := 0
  if y > rows
    y := rows
  if x > cols
    x := cols
      
  xloc := cursor[0] := x
  yloc := cursor[1] := y
  loc  := xloc + yloc * cols

  
PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    out("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      out(value/i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10



PUB hex(value, digits)

'' Print a hexadecimal number, specify number of digits

  repeat while digits > 8
    out("0")
    digits--

  value <<= (8 - digits) << 2

  repeat digits
    out(lookupz((value <-= 4) & $f : "0".."9", "A".."F"))

          
PUB out(c) | i, j, k, l, len, temp
'' Print a character
''
''  $09 = tab
''  $0D = return -> CR/LF
''  $20..$7E = display character
''  $C0   left arrow
''  $C1 = right arrow
''  $C2 = up arrow
''  $C3 = down arrow
''  $C4 = home key - go to beginning of line
''  $C5 = end key - go past last char on line
''  $C6 = page up key - skip this key
''  $C7 = page down key - skip this key
''  $C8 = backspace key
''  $C9 = delete key
''  $CA = clear from cursor to end of line 
''  $CB = insert line
''  $CC = delete line
''  $CD = cursor visible
''  $CE = cursor not visible
''  $CF = character invers
''  $D0 = character not invers
''  $D1 = save cursor
''  $D2 = restore cursor
''  $D3 = clear from cursor to end of screen 
''  $D4 = clear from start of screen to cursor
''  $D5 = Erase all of the display
''  $7F = skip

  case c

    $08:                        ' BS, cursor left
      if xloc > 0
       loc--                    ' screenpointer - 1
       xloc--                   ' cursorpointer - 1
       cursor[0] := xloc        ' set new cursor position
 
    $09:                        ' tab command
      repeat
        out( $20)               ' recursive call to out( )
      while xloc & 3            ' tab to multiples of 4
      
    $0A:                        ' LF, return to start of new line
     if yloc == rows-1          ' last column
        i := @screen            ' screen address
        j := i + cols           ' next column
        len := (chars - cols)/4 
        longmove(@screen, j, len)' shift screen up one line
 
       'clear last line
        j := i + (rows-1)*cols  ' row start position in pointer
        longfill(j,$20202020,cols/4) ' fill line with space
        cursor[1] := rows-1     ' new cursor position
        cursor[0] := xloc       ' set cursor     
        loc := xloc + (rows-1)*cols ' new output position

     else                       ' rows 0-38
        yloc++
        cursor[1] := yloc       ' new cursor position
        loc := xloc + yloc*cols ' output position

    $0D:                        ' CR, home key - move to 1st char of line
      xloc := cursor[0] := 0
      loc := xloc + yloc*cols


    ' normal output  
    $20..$BF:                   ' character                     //7E
      if inverse == 1           ' character visible invers
        c := c | %1_0000000
      screen[ loc++] := c       ' output the character

      if ++xloc == cols         ' bump the xloc
        xloc := 0
        yloc++
        if yloc == rows                                                               
          yloc := loc := 0

      if loc == chars           ' wrap if needed
        loc := xloc := yloc := cursor[0] := cursor[1] := 0
      else
        cursor[0] := xloc
        cursor[1] := yloc
        

    $C0:   ' left arrow
      if loc                    ' skip this if at upper left screen
        loc--
        if xloc
          xloc--
        else
          xloc := cols - 1
          yloc--
        cursor[0] := xloc
        cursor[1] := yloc


    $C1:   ' right arrow
      if loc <> chars - 1       'skip if at lower right of screen
        loc++
        if xloc <> cols - 1
          xloc++
        else
          xloc := 0
          yloc++
        cursor[0] := xloc
        cursor[1] := yloc


    $C2:   ' up arrow
      cursloc( xloc, yloc - 1)

    $C3:   ' down arrow
      cursloc( xloc, yloc + 1)

    $C4:   ' home key - move to 1st char of line
      xloc := cursor[0] := 0
      loc  := xloc + yloc*cols

    $C5:   ' end key - move to last char of line
      if xloc <> cols - 1
        repeat xloc from cols - 1 to 0
         loc := xloc + yloc*cols
         if screen[loc] <> $20  ' continue until first non-space char
           if xloc <> cols - 1
             xloc++               ' move past non-blank char
             loc++
           quit

        cursor[0] := xloc       ' loc is already reset from above

    $C8:   ' backspace
      if loc                    ' skip if at upper left of screen
        if xloc                 ' do 'else' if at start of line
          xloc--                ' xloc left one space
          loc--

          i := @screen          ' calculate
          i += xloc + yloc*cols ' destination for shift left one
          bytemove(i, i+1, cols - xloc - 1)
          screen[cols - 1 + yloc*cols] := $20

        else                    ' here if xloc == 0
          if screen[loc-1] == $20   ' last char on prev line
            yloc--

            i := @screen          ' calculate
            i += loc - 1          ' destination for shift left one

            repeat while screen[--loc] == $20
              bytemove(i, i+1, cols) ' move one row's worth of chars
              i--                 ' dec "i" to correspond to --loc

              screen[loc + cols] := $20  ' clear old char

              if ++xloc == cols   ' use xloc as counter here, 0..., don't move > 1 row
                loc--             ' make as if loc had been bumped above B4 we quit
                quit

            loc++                 ' bump loc to space char
            xloc := loc - yloc*cols ' re-calculate xloc from loc and yloc

        cursor[0] := xloc         ' reset cursor loc
        cursor[1] := yloc

    $C9:   ' delete
      if xloc == cols - 1
        screen[loc] := $20        ' if at last char on line, clear it and exit

      else
        repeat i from xloc to cols - 2
          j := i + yloc*cols
          screen[j] := screen[j+1]

       screen[j+1] := $20       ' clear last char on line after shift left

    $CA:   ' clear from cursor to end of line
      xloc := cursor[0]         ' column positio
      yloc := cursor[1]         ' row position
      repeat i from xloc to cols-1
        j := i + yloc*cols
        screen[j] := $20

    $CB:   ' insert line
      yloc := cursor[1]         ' row position
      i := @screen              ' screen addr
        
      if yloc < rows-1          ' rows 0-39
        j   := i + yloc*cols    ' yloc addr
        k   := j + cols         ' next line
        len := (chars - (yloc+1)*cols)/4
        longmove(k, j, len)     ' shift screen down   
        longfill(j,$20202020,cols/4) ' delete line      

      
      if yloc == rows-1         ' rows 39
        j := i + yloc*cols
        longfill(j,$20202020,cols/4) 'delete line 
       
        
    $CC:   ' delete line
      i    := @screen           ' screen address
      yloc := cursor[1]         ' row position
      i    := i + yloc*cols     ' row start position in pointer
      longfill(i,$20202020,cols/4) 'fill line with space

      
    $CD:   ' cursor 0 visible and underscore with slow blink
      cursor[2] := def_cursor     

    $CE:   ' cursor 0 not visible
      cursor[2] := %100      

    $CF:   ' set inverse flag
      inverse := 1           

    $D0:   ' reset inverse flag
      inverse := 0             

    $D1:   ' save cursor
        bytemove( @save_cursor, @cursor, 6)
      
    $D2:   ' restore cursor
        bytemove( @cursor, @save_cursor, 6)
        cursloc( cursor[0], cursor[1])

    $D3:   ' clear from cursor to end of screen
        i := chars - loc
        bytefill( @screen + loc, $20, i)   

    $D4:   ' clear from start of screen to cursor
        bytefill( @screen, $20, loc)   

    $D5:   ' Erase all of the display
        cls



PUB str(string_ptr)

'' Print a zero terminated string

  repeat strsize( string_ptr)
    out( byte[string_ptr++])